@@ -30,14 +30,16 @@
 
 #include <stdlib.h>
 #include <errno.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <string.h>
 
 #include "ruby.h"
 #include "ox.h"
+#include "sax.h"
 
 /* maximum to allocate on the stack, arbitrary limit */
-#define SMALL_XML	65536
+#define SMALL_XML	4096
 
 typedef struct _YesNoOpt {
     VALUE	sym;
@@ -48,12 +50,15 @@ void Init_ox();
 
 VALUE	 Ox = Qnil;
 
+ID	ox_at_column_id;
 ID	ox_at_content_id;
 ID	ox_at_id;
+ID	ox_at_line_id;
 ID	ox_at_value_id;
 ID	ox_attr_id;
 ID	ox_attr_value_id;
 ID	ox_attributes_id;
+ID	ox_attrs_done_id;
 ID	ox_beg_id;
 ID	ox_cdata_id;
 ID	ox_comment_id;
@@ -75,6 +80,7 @@ ID	ox_local_id;
 ID	ox_mesg_id;
 ID	ox_message_id;
 ID	ox_nodes_id;
+ID	ox_new_id;
 ID	ox_num_id;
 ID	ox_parse_id;
 ID	ox_read_id;
@@ -85,8 +91,8 @@ ID	ox_text_id;
 ID	ox_to_c_id;
 ID	ox_to_s_id;
 ID	ox_to_sym_id;
-ID	ox_tv_sec_id;
 ID	ox_tv_nsec_id;
+ID	ox_tv_sec_id;
 ID	ox_tv_usec_id;
 ID	ox_value_id;
 
@@ -97,6 +103,7 @@ VALUE	ox_zero_fixnum;
 
 VALUE	ox_arg_error_class;
 VALUE	ox_bag_clas;
+VALUE	ox_bigdecimal_class;
 VALUE	ox_cdata_clas;
 VALUE	ox_comment_clas;
 VALUE	ox_date_class;
@@ -126,8 +133,9 @@ static VALUE	object_sym;
 static VALUE	opt_format_sym;
 static VALUE	optimized_sym;
 static VALUE	strict_sym;
-static VALUE	strict_sym;
+static VALUE	smart_sym;
 static VALUE	symbolize_keys_sym;
+static VALUE	symbolize_sym;
 static VALUE	tolerant_sym;
 static VALUE	trace_sym;
 static VALUE	with_dtd_sym;
@@ -233,6 +241,9 @@ defuse_bom(char *xml, Options options) {
  * - effort: [:strict|:tolerant|:auto_define] set the tolerance level for loading
  * - symbolize_keys: [true|false|nil] symbolize element attribute keys or leave as Strings
  * @return [Hash] all current option settings.
+ *
+ * Note that an indent of less than zero will result in a tight one line output
+ * unless the text in the XML fields contain new line characters.
  */
 static VALUE
 get_def_opts(VALUE self) {
@@ -304,7 +315,7 @@ set_def_opts(VALUE self, VALUE opts) {
     } else {
 	Check_Type(v, T_STRING);
 	strncpy(ox_default_options.encoding, StringValuePtr(v), sizeof(ox_default_options.encoding) - 1);
-#ifdef HAVE_RUBY_ENCODING_H
+#if HAS_ENCODING_SUPPORT
 	ox_default_options.rb_enc = rb_enc_find(ox_default_options.encoding);
 #elif HAS_PRIVATE_ENCODING
 	ox_default_options.rb_enc = rb_str_new2(ox_default_options.encoding);
@@ -380,21 +391,33 @@ to_obj(VALUE self, VALUE ruby_xml) {
     size_t		len;
     VALUE		obj;
     struct _Options	options = ox_default_options;
+    struct _Err		err;
 
+    err_init(&err);
     Check_Type(ruby_xml, T_STRING);
     /* the xml string gets modified so make a copy of it */
     len = RSTRING_LEN(ruby_xml) + 1;
+    x = defuse_bom(StringValuePtr(ruby_xml), &options);
     if (SMALL_XML < len) {
 	xml = ALLOC_N(char, len);
     } else {
 	xml = ALLOCA_N(char, len);
     }
-    memcpy(xml, StringValuePtr(ruby_xml), len);
-    x = defuse_bom(xml, &options);
-    obj = ox_parse(x, ox_obj_callbacks, 0, &options);
+    memcpy(xml, x, len);
+#if HAS_GC_GUARD
+    rb_gc_disable();
+#endif
+    obj = ox_parse(xml, ox_obj_callbacks, 0, &options, &err);
     if (SMALL_XML < len) {
 	xfree(xml);
     }
+#if HAS_GC_GUARD
+    RB_GC_GUARD(obj);
+    rb_gc_enable();
+#endif
+    if (err_has(&err)) {
+	ox_err_raise(&err);
+    }
     return obj;
 }
 
@@ -411,29 +434,34 @@ to_gen(VALUE self, VALUE ruby_xml) {
     size_t		len;
     VALUE		obj;
     struct _Options	options = ox_default_options;
+    struct _Err		err;
 
+    err_init(&err);
     Check_Type(ruby_xml, T_STRING);
     /* the xml string gets modified so make a copy of it */
     len = RSTRING_LEN(ruby_xml) + 1;
+    x = defuse_bom(StringValuePtr(ruby_xml), &options);
     if (SMALL_XML < len) {
 	xml = ALLOC_N(char, len);
     } else {
 	xml = ALLOCA_N(char, len);
     }
-    memcpy(xml, StringValuePtr(ruby_xml), len);
-    x = defuse_bom(xml, &options);
-    obj = ox_parse(x, ox_gen_callbacks, 0, &options);
+    memcpy(xml, x, len);
+    obj = ox_parse(xml, ox_gen_callbacks, 0, &options, &err);
     if (SMALL_XML < len) {
 	xfree(xml);
     }
+    if (err_has(&err)) {
+	ox_err_raise(&err);
+    }
     return obj;
 }
 
 static VALUE
-load(char *xml, int argc, VALUE *argv, VALUE self, VALUE encoding) {
+load(char *xml, int argc, VALUE *argv, VALUE self, VALUE encoding, Err err) {
     VALUE		obj;
     struct _Options	options = ox_default_options;
-    
+
     if (1 == argc && rb_cHash == rb_obj_class(*argv)) {
 	VALUE	h = *argv;
 	VALUE	v;
@@ -470,7 +498,7 @@ load(char *xml, int argc, VALUE *argv, VALUE self, VALUE encoding) {
 	    options.sym_keys = (Qfalse == v) ? No : Yes;
 	}
     }
-#ifdef HAVE_RUBY_ENCODING_H
+#if HAS_ENCODING_SUPPORT
     if ('\0' == *options.encoding) {
 	if (Qnil != encoding) {
 	    options.rb_enc = rb_enc_from_index(rb_enc_get_index(encoding));
@@ -495,19 +523,26 @@ load(char *xml, int argc, VALUE *argv, VALUE self, VALUE encoding) {
     xml = defuse_bom(xml, &options);
     switch (options.mode) {
     case ObjMode:
-	obj = ox_parse(xml, ox_obj_callbacks, 0, &options);
+#if HAS_GC_GUARD
+	rb_gc_disable();
+#endif
+	obj = ox_parse(xml, ox_obj_callbacks, 0, &options, err);
+#if HAS_GC_GUARD
+	RB_GC_GUARD(obj);
+	rb_gc_enable();
+#endif
 	break;
     case GenMode:
-	obj = ox_parse(xml, ox_gen_callbacks, 0, &options);
+	obj = ox_parse(xml, ox_gen_callbacks, 0, &options, err);
 	break;
     case LimMode:
-	obj = ox_parse(xml, ox_limited_callbacks, 0, &options);
+	obj = ox_parse(xml, ox_limited_callbacks, 0, &options, err);
 	break;
     case NoMode:
-	obj = ox_parse(xml, ox_nomode_callbacks, 0, &options);
+	obj = ox_parse(xml, ox_nomode_callbacks, 0, &options, err);
 	break;
     default:
-	obj = ox_parse(xml, ox_gen_callbacks, 0, &options);
+	obj = ox_parse(xml, ox_gen_callbacks, 0, &options, err);
 	break;
     }
     return obj;
@@ -537,7 +572,9 @@ load_str(int argc, VALUE *argv, VALUE self) {
     size_t	len;
     VALUE	obj;
     VALUE	encoding;
-    
+    struct _Err	err;
+
+    err_init(&err);
     Check_Type(*argv, T_STRING);
     /* the xml string gets modified so make a copy of it */
     len = RSTRING_LEN(*argv) + 1;
@@ -546,18 +583,25 @@ load_str(int argc, VALUE *argv, VALUE self) {
     } else {
 	xml = ALLOCA_N(char, len);
     }
-#ifdef HAVE_RUBY_ENCODING_H
+#if HAS_ENCODING_SUPPORT
+#ifdef MACRUBY_RUBY
+    encoding = rb_funcall(*argv, rb_intern("encoding"), 0);
+#else
     encoding = rb_obj_encoding(*argv);
+#endif
 #elif HAS_PRIVATE_ENCODING
     encoding = rb_funcall(*argv, rb_intern("encoding"), 0);
 #else
     encoding = Qnil;
 #endif
     memcpy(xml, StringValuePtr(*argv), len);
-    obj = load(xml, argc - 1, argv + 1, self, encoding);
+    obj = load(xml, argc - 1, argv + 1, self, encoding, &err);
     if (SMALL_XML < len) {
 	xfree(xml);
     }
+    if (err_has(&err)) {
+	ox_err_raise(&err);
+    }
     return obj;
 }
 
@@ -586,7 +630,9 @@ load_file(int argc, VALUE *argv, VALUE self) {
     FILE	*f;
     size_t	len;
     VALUE	obj;
-    
+    struct _Err	err;
+
+    err_init(&err);
     Check_Type(*argv, T_STRING);
     path = StringValuePtr(*argv);
     if (0 == (f = fopen(path, "r"))) {
@@ -601,15 +647,19 @@ load_file(int argc, VALUE *argv, VALUE self) {
     }
     fseek(f, 0, SEEK_SET);
     if (len != fread(xml, 1, len, f)) {
-	fclose(f);
-	rb_raise(rb_eLoadError, "Failed to read %ld bytes from %s.\n", (long)len, path);
+	ox_err_set(&err, rb_eLoadError, "Failed to read %ld bytes from %s.\n", (long)len, path);
+	obj = Qnil;
+    } else {
+	xml[len] = '\0';
+	obj = load(xml, argc - 1, argv + 1, self, Qnil, &err);
     }
     fclose(f);
-    xml[len] = '\0';
-    obj = load(xml, argc - 1, argv + 1, self, Qnil);
     if (SMALL_XML < len) {
 	xfree(xml);
     }
+    if (err_has(&err)) {
+	ox_err_raise(&err);
+    }
     return obj;
 }
 
@@ -620,11 +670,17 @@ load_file(int argc, VALUE *argv, VALUE self) {
  * @param [Ox::Sax] handler SAX (responds to OX::Sax methods) like handler
  * @param [IO|String] io IO Object to read from
  * @param [Hash] options parse options
- * @param [true|false] :convert_special flag indicating special special characters like &lt; are converted
+ * @param [true|false] :convert_special flag indicating special characters like &lt; are converted
+ * @param [true|false] :symbolize flag indicating the parser symbolize element and attribute names
+ * @param [true|false] :smart flag indicating the parser use hints if available (use with html)
  */
 static VALUE
 sax_parse(int argc, VALUE *argv, VALUE self) {
-    int convert = 0;
+    struct _SaxOptions	options;
+
+    options.symbolize = 1;
+    options.convert_special = 0;
+    options.smart = 0;
 
     if (argc < 2) {
 	rb_raise(ox_parse_error_class, "Wrong number of arguments to sax_parse.\n");
@@ -634,10 +690,16 @@ sax_parse(int argc, VALUE *argv, VALUE self) {
 	VALUE	v;
 	
 	if (Qnil != (v = rb_hash_lookup(h, convert_special_sym))) {
-	    convert = (Qtrue == v);
+	    options.convert_special = (Qtrue == v);
+	}
+	if (Qnil != (v = rb_hash_lookup(h, smart_sym))) {
+	    options.smart = (Qtrue == v);
+	}
+	if (Qnil != (v = rb_hash_lookup(h, symbolize_sym))) {
+	    options.symbolize = (Qtrue == v);
 	}
     }
-    ox_sax_parse(argv[0], argv[1], convert);
+    ox_sax_parse(argv[0], argv[1], &options);
 
     return Qnil;
 }
@@ -713,6 +775,9 @@ parse_dump_options(VALUE ropts, Options copts) {
  * @param [:strict|:tolerant] :effort effort to use when an undumpable object (e.g., IO) is encountered, default: :strict
  *  - *:strict* - raise an NotImplementedError if an undumpable object is encountered
  *  - *:tolerant* - replaces undumplable objects with nil
+ *
+ * Note that an indent of less than zero will result in a tight one line output
+ * unless the text in the XML fields contain new line characters.
  */
 static VALUE
 dump(int argc, VALUE *argv, VALUE self) {
@@ -753,6 +818,9 @@ dump(int argc, VALUE *argv, VALUE self) {
  * @param [:strict|:tolerant] :effort effort to use when an undumpable object (e.g., IO) is encountered, default: :strict
  *  - *:strict* - raise an NotImplementedError if an undumpable object is encountered
  *  - *:tolerant* - replaces undumplable objects with nil
+ *
+ * Note that an indent of less than zero will result in a tight one line output
+ * unless the text in the XML fields contain new line characters.
  */
 static VALUE
 to_file(int argc, VALUE *argv, VALUE self) {
@@ -802,25 +870,29 @@ void Init_ox() {
 
     rb_require("time");
     rb_require("date");
+    rb_require("bigdecimal");
     rb_require("stringio");
 
+    ox_at_column_id = rb_intern("@column");
+    ox_at_content_id = rb_intern("@content");
     ox_at_id = rb_intern("at");
+    ox_at_line_id = rb_intern("@line");
     ox_at_value_id = rb_intern("@value");
     ox_attr_id = rb_intern("attr");
     ox_attr_value_id = rb_intern("attr_value");
     ox_attributes_id = rb_intern("@attributes");
+    ox_attrs_done_id = rb_intern("attrs_done");
     ox_beg_id = rb_intern("@beg");
     ox_cdata_id = rb_intern("cdata");
     ox_comment_id = rb_intern("comment");
-    ox_at_content_id = rb_intern("@content");
     ox_den_id = rb_intern("@den");
     ox_doctype_id = rb_intern("doctype");
-    ox_external_encoding_id = rb_intern("external_encoding");
     ox_end_element_id = rb_intern("end_element");
     ox_end_id = rb_intern("@end");
     ox_end_instruct_id = rb_intern("end_instruct");
     ox_error_id = rb_intern("error");
     ox_excl_id = rb_intern("@excl");
+    ox_external_encoding_id = rb_intern("external_encoding");
     ox_fileno_id = rb_intern("fileno");
     ox_force_encoding_id = rb_intern("force_encoding");
     ox_inspect_id = rb_intern("inspect");
@@ -831,20 +903,34 @@ void Init_ox() {
     ox_mesg_id = rb_intern("mesg");
     ox_message_id = rb_intern("message");
     ox_nodes_id = rb_intern("@nodes");
+    ox_new_id = rb_intern("new");
     ox_num_id = rb_intern("@num");
     ox_parse_id = rb_intern("parse");
-    ox_readpartial_id = rb_intern("readpartial");
     ox_read_id = rb_intern("read");
+    ox_readpartial_id = rb_intern("readpartial");
     ox_start_element_id = rb_intern("start_element");
     ox_string_id = rb_intern("string");
     ox_text_id = rb_intern("text");
-    ox_value_id = rb_intern("value");
     ox_to_c_id = rb_intern("to_c");
     ox_to_s_id = rb_intern("to_s");
     ox_to_sym_id = rb_intern("to_sym");
-    ox_tv_sec_id = rb_intern("tv_sec");
     ox_tv_nsec_id = rb_intern("tv_nsec");
+    ox_tv_sec_id = rb_intern("tv_sec");
     ox_tv_usec_id = rb_intern("tv_usec");
+    ox_value_id = rb_intern("value");
+
+    rb_require("ox/version");
+    rb_require("ox/error");
+    rb_require("ox/hasattrs");
+    rb_require("ox/node");
+    rb_require("ox/comment");
+    rb_require("ox/instruct");
+    rb_require("ox/cdata");
+    rb_require("ox/doctype");
+    rb_require("ox/element");
+    rb_require("ox/document");
+    rb_require("ox/bag");
+    rb_require("ox/sax");
 
     ox_time_class = rb_const_get(rb_cObject, rb_intern("Time"));
     ox_date_class = rb_const_get(rb_cObject, rb_intern("Date"));
@@ -852,6 +938,7 @@ void Init_ox() {
     ox_arg_error_class = rb_const_get_at(Ox, rb_intern("ArgError"));
     ox_struct_class = rb_const_get(rb_cObject, rb_intern("Struct"));
     ox_stringio_class = rb_const_get(rb_cObject, rb_intern("StringIO"));
+    ox_bigdecimal_class = rb_const_get(rb_cObject, rb_intern("BigDecimal"));
 
     auto_define_sym = ID2SYM(rb_intern("auto_define"));		rb_gc_register_address(&auto_define_sym);
     auto_sym = ID2SYM(rb_intern("auto"));			rb_gc_register_address(&auto_sym);
@@ -866,8 +953,10 @@ void Init_ox() {
     opt_format_sym = ID2SYM(rb_intern("opt_format"));		rb_gc_register_address(&opt_format_sym);
     optimized_sym = ID2SYM(rb_intern("optimized"));		rb_gc_register_address(&optimized_sym);
     ox_encoding_sym = ID2SYM(rb_intern("encoding"));		rb_gc_register_address(&ox_encoding_sym);
+    smart_sym = ID2SYM(rb_intern("smart"));			rb_gc_register_address(&smart_sym);
     strict_sym = ID2SYM(rb_intern("strict"));			rb_gc_register_address(&strict_sym);
     symbolize_keys_sym = ID2SYM(rb_intern("symbolize_keys"));	rb_gc_register_address(&symbolize_keys_sym);
+    symbolize_sym = ID2SYM(rb_intern("symbolize"));		rb_gc_register_address(&symbolize_sym);
     tolerant_sym = ID2SYM(rb_intern("tolerant"));		rb_gc_register_address(&tolerant_sym);
     trace_sym = ID2SYM(rb_intern("trace"));			rb_gc_register_address(&trace_sym);
     with_dtd_sym = ID2SYM(rb_intern("with_dtd"));		rb_gc_register_address(&with_dtd_sym);
@@ -914,5 +1003,8 @@ _ox_raise_error(const char *msg, const char *xml, const char *current, const cha
 	    xline++;
 	}
     }
+#if HAS_GC_GUARD
+    rb_gc_enable();
+#endif
     rb_raise(ox_parse_error_class, "%s at line %d, column %d [%s:%d]\n", msg, xline, col, file, line);
 }
