@@ -35,11 +35,15 @@
 
 #include "ruby.h"
 #include "ox.h"
+#include "err.h"
+#include "attr.h"
+#include "helper.h"
+#include "special.h"
 
 static void	read_instruction(PInfo pi);
 static void	read_doctype(PInfo pi);
 static void	read_comment(PInfo pi);
-static void	read_element(PInfo pi);
+static char*	read_element(PInfo pi);
 static void	read_text(PInfo pi);
 /*static void	  read_reduced_text(PInfo pi); */
 static void	read_cdata(PInfo pi);
@@ -47,7 +51,6 @@ static char*	read_name_token(PInfo pi);
 static char*	read_quoted_value(PInfo pi);
 static char*	read_hex_uint64(char *b, uint64_t *up);
 static char*	read_10_uint64(char *b, uint64_t *up);
-static char*	ucs_to_utf8_chars(char *text, uint64_t u);
 static char*	read_coded_chars(PInfo pi, char *text);
 static void	next_non_white(PInfo pi);
 static int	collapse_special(PInfo pi, char *str);
@@ -99,20 +102,22 @@ next_white(PInfo pi) {
 }
 
 VALUE
-ox_parse(char *xml, ParseCallbacks pcb, char **endp, Options options) {
+ox_parse(char *xml, ParseCallbacks pcb, char **endp, Options options, Err err) {
     struct _PInfo	pi;
     int			body_read = 0;
 
     if (0 == xml) {
-	raise_error("Invalid arg, xml string can not be null", xml, 0);
+	set_error(err, "Invalid arg, xml string can not be null", xml, 0);
+	return Qnil;
     }
     if (DEBUG <= options->trace) {
 	printf("Parsing xml:\n%s\n", xml);
     }
     /* initialize parse info */
+    helper_stack_init(&pi.helpers);
+    err_init(&pi.err);
     pi.str = xml;
     pi.s = xml;
-    pi.h = 0;
     pi.pcb = pcb;
     pi.obj = Qnil;
     pi.circ_array = 0;
@@ -127,7 +132,9 @@ ox_parse(char *xml, ParseCallbacks pcb, char **endp, Options options) {
 	    break;
 	}
 	if ('<' != *pi.s) {		/* all top level entities start with < */
-	    raise_error("invalid format, expected <", pi.str, pi.s);
+	    set_error(err, "invalid format, expected <", pi.str, pi.s);
+	    helper_stack_cleanup(&pi.helpers);
+	    return Qnil;
 	}
 	pi.s++;		/* past < */
 	switch (*pi.s) {
@@ -138,30 +145,44 @@ ox_parse(char *xml, ParseCallbacks pcb, char **endp, Options options) {
 	case '!':	/* comment or doctype */
 	    pi.s++;
 	    if ('\0' == *pi.s) {
-		raise_error("invalid format, DOCTYPE or comment not terminated", pi.str, pi.s);
+		set_error(err, "invalid format, DOCTYPE or comment not terminated", pi.str, pi.s);
+		helper_stack_cleanup(&pi.helpers);
+		return Qnil;
 	    } else if ('-' == *pi.s) {
 		pi.s++;	/* skip - */
 		if ('-' != *pi.s) {
-		    raise_error("invalid format, bad comment format", pi.str, pi.s);
+		    set_error(err, "invalid format, bad comment format", pi.str, pi.s);
+		    helper_stack_cleanup(&pi.helpers);
+		    return Qnil;
 		} else {
 		    pi.s++;	/* skip second - */
 		    read_comment(&pi);
 		}
-	    } else if (0 == strncmp("DOCTYPE", pi.s, 7)) {
+	    } else if ((TolerantEffort == options->effort) ? 0 == strncasecmp("DOCTYPE", pi.s, 7) : 0 == strncmp("DOCTYPE", pi.s, 7)) {
 		pi.s += 7;
 		read_doctype(&pi);
 	    } else {
-		raise_error("invalid format, DOCTYPE or comment expected", pi.str, pi.s);
+		set_error(err, "invalid format, DOCTYPE or comment expected", pi.str, pi.s);
+		helper_stack_cleanup(&pi.helpers);
+		return Qnil;
 	    }
 	    break;
 	case '\0':
-	    raise_error("invalid format, document not terminated", pi.str, pi.s);
+	    set_error(err, "invalid format, document not terminated", pi.str, pi.s);
+	    helper_stack_cleanup(&pi.helpers);
+	    return Qnil;
 	default:
 	    read_element(&pi);
 	    body_read = 1;
 	    break;
 	}
+	if (err_has(&pi.err)) {
+	    *err = pi.err;
+	    helper_stack_cleanup(&pi.helpers);
+	    return Qnil;
+	}
     }
+    helper_stack_cleanup(&pi.helpers);
     return pi.obj;
 }
 
@@ -191,31 +212,42 @@ gather_content(const char *src, char *content, size_t len) {
 static void
 read_instruction(PInfo pi) {
     char		content[1024];
-    struct _Attr	attrs[MAX_ATTRS + 1];
-    Attr		a = attrs;
+    struct _AttrStack	attrs;
+    char		*attr_name;
+    char		*attr_value;
     char		*target;
     char		*end;
     char		c;
     char		*cend;
     int			attrs_ok = 1;
 
+
     *content = '\0';
-    memset(attrs, 0, sizeof(attrs));
-    target = read_name_token(pi);
+    attr_stack_init(&attrs);
+    if (0 == (target = read_name_token(pi))) {
+	return;
+    }
     end = pi->s;
     if (0 == (cend = gather_content(pi->s, content, sizeof(content) - 1))) {
-	raise_error("processing instruction content too large or not terminated", pi->str, pi->s);
+	set_error(&pi->err, "processing instruction content too large or not terminated", pi->str, pi->s);
+	return;
     }
     next_non_white(pi);
     c = *pi->s;
     *end = '\0'; /* terminate name */
     if ('?' != c) {
-	while ('?' != *pi->s) {
+	while ('?' != c) {
+	    pi->last = 0;
 	    if ('\0' == *pi->s) {
-		raise_error("invalid format, processing instruction not terminated", pi->str, pi->s);
+		attr_stack_cleanup(&attrs);
+		set_error(&pi->err, "invalid format, processing instruction not terminated", pi->str, pi->s);
+		return;
 	    }
 	    next_non_white(pi);
-	    a->name = read_name_token(pi);
+	    if (0 == (attr_name = read_name_token(pi))) {
+		attr_stack_cleanup(&attrs);
+		return;
+	    }
 	    end = pi->s;
 	    next_non_white(pi);
 	    if ('=' != *pi->s++) {
@@ -225,13 +257,17 @@ read_instruction(PInfo pi) {
 	    *end = '\0'; /* terminate name */
 	    /* read value */
 	    next_non_white(pi);
-	    a->value = read_quoted_value(pi);
-	    a++;
-	    if (MAX_ATTRS <= (a - attrs)) {
-		attrs_ok = 0;
-		break;
+	    if (0 == (attr_value = read_quoted_value(pi))) {
+		attr_stack_cleanup(&attrs);
+		return;
 	    }
+	    attr_stack_push(&attrs, attr_name, attr_value);
 	    next_non_white(pi);
+	    if ('\0' == pi->last) {
+		c = *pi->s;
+	    } else {
+		c = pi->last;
+	    }
 	}
 	if ('?' == *pi->s) {
 	    pi->s++;
@@ -241,45 +277,77 @@ read_instruction(PInfo pi) {
     }
     if (attrs_ok) {
 	if ('>' != *pi->s++) {
-	    raise_error("invalid format, processing instruction not terminated", pi->str, pi->s);
+	    attr_stack_cleanup(&attrs);
+	    set_error(&pi->err, "invalid format, processing instruction not terminated", pi->str, pi->s);
+	    return;
 	}
     } else {
 	pi->s = cend + 1;
     }
     if (0 != pi->pcb->instruct) {
 	if (attrs_ok) {
-	    pi->pcb->instruct(pi, target, attrs, 0);
+	    pi->pcb->instruct(pi, target, attrs.head, 0);
 	} else {
-	    pi->pcb->instruct(pi, target, attrs, content);
+	    pi->pcb->instruct(pi, target, attrs.head, content);
+	}
+    }
+    attr_stack_cleanup(&attrs);
+}
+
+static void
+read_delimited(PInfo pi, char end) {
+    char	c;
+
+    if ('"' == end || '\'' == end) {
+	for (c = *pi->s++; end != c; c = *pi->s++) {
+	    if ('\0' == c) {
+		set_error(&pi->err, "invalid format, dectype not terminated", pi->str, pi->s);
+		return;
+	    }
+	}
+    } else {
+	while (1) {
+	    c = *pi->s++;
+	    if (end == c) {
+		return;
+	    }
+	    switch (c) {
+	    case '\0':
+		set_error(&pi->err, "invalid format, dectype not terminated", pi->str, pi->s);
+		return;
+	    case '"':
+		read_delimited(pi, c);
+		break;
+	    case '\'':
+		read_delimited(pi, c);
+		break;
+	    case '[':
+		read_delimited(pi, ']');
+		break;
+	    case '<':
+		read_delimited(pi, '>');
+		break;
+	    default:
+		break;
+	    }
 	}
     }
 }
 
 /* Entered after the "<!DOCTYPE" sequence plus the first character after
- * that. Ready to read the rest. Returns error code.
+ * that. Ready to read the rest.
  */
 static void
 read_doctype(PInfo pi) {
     char	*docType;
-    int		depth = 1;
-    char	c;
 
     next_non_white(pi);
     docType = pi->s;
-    while (1) {
-	c = *pi->s++;
-	if ('\0' == c) {
-	    raise_error("invalid format, prolog not terminated", pi->str, pi->s);
-	} else if ('<' == c) {
-	    depth++;
-	} else if ('>' == c) {
-	    depth--;
-	    if (0 == depth) {	/* done, at the end */
-		pi->s--;
-		break;
-	    }
-	}
+    read_delimited(pi, '>');
+    if (err_has(&pi->err)) {
+	return;
     }
+    pi->s--;
     *pi->s = '\0';
     pi->s++;
     if (0 != pi->pcb->add_doctype) {
@@ -300,7 +368,8 @@ read_comment(PInfo pi) {
     comment = pi->s;
     end = strstr(pi->s, "-->");
     if (0 == end) {
-	raise_error("invalid format, comment not terminated", pi->str, pi->s);
+	set_error(&pi->err, "invalid format, comment not terminated", pi->str, pi->s);
+	return;
     }
     for (s = end - 1; pi->s < s && !done; s--) {
 	switch(*s) {
@@ -326,10 +395,11 @@ read_comment(PInfo pi) {
 /* Entered after the '<' and the first character after that. Returns status
  * code.
  */
-static void
+static char*
 read_element(PInfo pi) {
-    struct _Attr	attrs[MAX_ATTRS];
-    Attr		ap = attrs;
+    struct _AttrStack	attrs;
+    const char		*attr_name;
+    const char		*attr_value;
     char		*name;
     char		*ename;
     char		*end;
@@ -338,7 +408,10 @@ read_element(PInfo pi) {
     int			hasChildren = 0;
     int			done = 0;
 
-    ename = read_name_token(pi);
+    attr_stack_init(&attrs);
+    if (0 == (ename = read_name_token(pi))) {
+	return 0;
+    }
     end = pi->s;
     elen = end - ename;
     next_non_white(pi);
@@ -349,14 +422,16 @@ read_element(PInfo pi) {
 	pi->s++;
 	if ('>' != *pi->s) {
 	    /*printf("*** '%s' ***\n", pi->s); */
-	    raise_error("invalid format, element not closed", pi->str, pi->s);
+	    attr_stack_cleanup(&attrs);
+	    set_error(&pi->err, "invalid format, element not closed", pi->str, pi->s);
+	    return 0;
 	}
 	pi->s++;	/* past > */
-	ap->name = 0;
-	pi->pcb->add_element(pi, ename, attrs, hasChildren);
+	pi->pcb->add_element(pi, ename, attrs.head, hasChildren);
 	pi->pcb->end_element(pi, ename);
 
-	return;
+	attr_stack_cleanup(&attrs);
+	return 0;
     }
     /* read attribute names until the close (/ or >) is reached */
     while (!done) {
@@ -364,54 +439,78 @@ read_element(PInfo pi) {
 	    next_non_white(pi);
 	    c = *pi->s;
 	}
+	pi->last = 0;
 	switch (c) {
 	case '\0':
-	    raise_error("invalid format, document not terminated", pi->str, pi->s);
+	    attr_stack_cleanup(&attrs);
+	    set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+	    return 0;
 	case '/':
 	    /* Element with just attributes. */
 	    pi->s++;
 	    if ('>' != *pi->s) {
-		raise_error("invalid format, element not closed", pi->str, pi->s);
+		attr_stack_cleanup(&attrs);
+		set_error(&pi->err, "invalid format, element not closed", pi->str, pi->s);
+		return 0;
 	    }
 	    pi->s++;
-	    ap->name = 0;
-	    pi->pcb->add_element(pi, ename, attrs, hasChildren);
+	    pi->pcb->add_element(pi, ename, attrs.head, hasChildren);
 	    pi->pcb->end_element(pi, ename);
 
-	    return;
+	    attr_stack_cleanup(&attrs);
+	    return 0;
 	case '>':
 	    /* has either children or a value */
 	    pi->s++;
 	    hasChildren = 1;
 	    done = 1;
-	    ap->name = 0;
-	    pi->pcb->add_element(pi, ename, attrs, hasChildren);
+	    pi->pcb->add_element(pi, ename, attrs.head, hasChildren);
 	    break;
 	default:
 	    /* Attribute name so it's an element and the attribute will be */
 	    /* added to it. */
-	    ap->name = read_name_token(pi);
+	    if (0 == (attr_name = read_name_token(pi))) {
+		attr_stack_cleanup(&attrs);
+		return 0;
+	    }
 	    end = pi->s;
 	    next_non_white(pi);
 	    if ('=' != *pi->s++) {
-		raise_error("invalid format, no attribute value", pi->str, pi->s);
+		if (TolerantEffort == pi->options->effort) {
+		    pi->s--;
+		    pi->last = *pi->s;
+		    *end = '\0'; /* terminate name */
+		    attr_value = "";
+		    attr_stack_push(&attrs, attr_name, attr_value);
+		    break;
+		} else {
+		    attr_stack_cleanup(&attrs);
+		    set_error(&pi->err, "invalid format, no attribute value", pi->str, pi->s);
+		    return 0;
+		}
 	    }
 	    *end = '\0'; /* terminate name */
 	    /* read value */
 	    next_non_white(pi);
-	    ap->value = read_quoted_value(pi);
-	    if (0 != strchr(ap->value, '&')) {
-		if (0 != collapse_special(pi, (char*)ap->value)) {
-		    raise_error("invalid format, special character does not end with a semicolon", pi->str, pi->s);
+	    if (0 == (attr_value = read_quoted_value(pi))) {
+		return 0;
+	    }
+	    if (0 != strchr(attr_value, '&')) {
+		if (0 != collapse_special(pi, (char*)attr_value)) {
+		    attr_stack_cleanup(&attrs);
+		    set_error(&pi->err, "invalid format, special character does not end with a semicolon", pi->str, pi->s);
+		    return 0;
 		}
 	    }
-	    ap++;
-	    if (MAX_ATTRS <= (ap - attrs)) {
-		raise_error("too many attributes", pi->str, pi->s);
-	    }
+	    attr_stack_push(&attrs, attr_name, attr_value);
 	    break;
 	}
-	c = '\0';
+	if ('\0' == pi->last) {
+	    c = '\0';
+	} else {
+	    c = pi->last;
+	    pi->last = '\0';
+	}
     }
     if (hasChildren) {
 	char	*start;
@@ -424,7 +523,9 @@ read_element(PInfo pi) {
 	    next_non_white(pi);
 	    c = *pi->s++;
 	    if ('\0' == c) {
-		raise_error("invalid format, document not terminated", pi->str, pi->s);
+		attr_stack_cleanup(&attrs);
+		set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+		return 0;
 	    }
 	    if ('<' == c) {
 		char	*slash;
@@ -435,11 +536,15 @@ read_element(PInfo pi) {
 		    if ('-' == *pi->s && '-' == *(pi->s + 1)) {
 			pi->s += 2;
 			read_comment(pi);
-		    } else if (0 == strncmp("[CDATA[", pi->s, 7)) {
+		    } else if ((TolerantEffort == pi->options->effort) ?
+			       0 == strncasecmp("[CDATA[", pi->s, 7) :
+			       0 == strncmp("[CDATA[", pi->s, 7)) {
 			pi->s += 7;
 			read_cdata(pi);
 		    } else {
-			raise_error("invalid format, invalid comment or CDATA format", pi->str, pi->s);
+			attr_stack_cleanup(&attrs);
+			set_error(&pi->err, "invalid format, invalid comment or CDATA format", pi->str, pi->s);
+			return 0;
 		    }
 		    break;
 		case '?':	/* processing instruction */
@@ -449,16 +554,28 @@ read_element(PInfo pi) {
 		case '/':
 		    slash = pi->s;
 		    pi->s++;
-		    name = read_name_token(pi);
+		    if (0 == (name = read_name_token(pi))) {
+			attr_stack_cleanup(&attrs);
+			return 0;
+		    }
 		    end = pi->s;
 		    next_non_white(pi);
 		    c = *pi->s;
 		    *end = '\0';
 		    if (0 != strcmp(name, ename)) {
-			raise_error("invalid format, elements overlap", pi->str, pi->s);
+			attr_stack_cleanup(&attrs);
+			if (TolerantEffort == pi->options->effort) {
+			    pi->pcb->end_element(pi, ename);
+			    return name;
+			} else {
+			    set_error(&pi->err, "invalid format, elements overlap", pi->str, pi->s);
+			    return 0;
+			}
 		    }
 		    if ('>' != c) {
-			raise_error("invalid format, element not closed", pi->str, pi->s);
+			attr_stack_cleanup(&attrs);
+			set_error(&pi->err, "invalid format, element not closed", pi->str, pi->s);
+			return 0;
 		    }
 		    if (first && start != slash - 1) {
 			/* some white space between start and here so add as text */
@@ -467,13 +584,33 @@ read_element(PInfo pi) {
 		    }
 		    pi->s++;
 		    pi->pcb->end_element(pi, ename);
-		    return;
+		    attr_stack_cleanup(&attrs);
+		    return 0;
 		case '\0':
-		    raise_error("invalid format, document not terminated", pi->str, pi->s);
+		    attr_stack_cleanup(&attrs);
+		    if (TolerantEffort == pi->options->effort) {
+			return 0;
+		    } else {
+			set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+			return 0;
+		    }
 		default:
 		    first = 0;
 		    /* a child element */
-		    read_element(pi);
+		    // Child closed with mismatched name.
+		    if (0 != (name = read_element(pi))) {
+			attr_stack_cleanup(&attrs);
+			if (0 == strcmp(name, ename)) {
+			    pi->s++;
+			    pi->pcb->end_element(pi, ename);
+			    return 0;
+			} else { // not the correct element yet
+			    pi->pcb->end_element(pi, ename);
+			    return name;
+			}
+		    } else if (err_has(&pi->err)) {
+			return 0;
+		    }
 		    break;
 		}
 	    } else {	/* read as TEXT */
@@ -489,11 +626,14 @@ read_element(PInfo pi) {
 		    /* close tag after text so treat as a value */
 		    pi->s += elen + 3;
 		    pi->pcb->end_element(pi, ename);
-		    return;
+		    attr_stack_cleanup(&attrs);
+		    return 0;
 		}
 	    }
 	}
     }
+    attr_stack_cleanup(&attrs);
+    return 0;
 }
 
 static void
@@ -513,7 +653,8 @@ read_text(PInfo pi) {
 	    pi->s--;
 	    break;
 	case '\0':
-	    raise_error("invalid format, document not terminated", pi->str, pi->s);
+	    set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+	    return;
 	default:
 	    if (end <= (b + (('&' == c) ? 7 : 0))) { /* extra 8 for special just in case it is sequence of bytes */
 		unsigned long	size;
@@ -533,7 +674,9 @@ read_text(PInfo pi) {
 		end = alloc_buf + size - 2;
 	    }
 	    if ('&' == c) {
-		b = read_coded_chars(pi, b);
+		if (0 == (b = read_coded_chars(pi, b))) {
+		    return;
+		}
 	    } else {
 		*b++ = c;
 	    }
@@ -575,7 +718,8 @@ read_reduced_text(PInfo pi) {
 	    pi->s--;
 	    break;
 	case '\0':
-	    raise_error("invalid format, document not terminated", pi->str, pi->s);
+	    set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+	    return;
 	default:
 	    if (end <= (b + spc + (('&' == c) ? 7 : 0))) { /* extra 8 for special just in case it is sequence of bytes */
 		unsigned long	size;
@@ -599,7 +743,9 @@ read_reduced_text(PInfo pi) {
 	    }
 	    spc = 0;
 	    if ('&' == c) {
-		b = read_coded_chars(pi, b);
+		if (0 == (b = read_coded_chars(pi, b))) {
+		    return;
+		}
 	    } else {
 		*b++ = c;
 	    }
@@ -636,7 +782,8 @@ read_name_token(PInfo pi) {
 	    return start;
 	case '\0':
 	    /* documents never terminate after a name token */
-	    raise_error("invalid format, document not terminated", pi->str, pi->s);
+	    set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+	    return 0;
 	    break; /* to avoid warnings */
 	default:
 	    break;
@@ -653,7 +800,8 @@ read_cdata(PInfo pi) {
     start = pi->s;
     end = strstr(pi->s, "]]>");
     if (end == 0) {
-	raise_error("invalid format, CDATA not terminated", pi->str, pi->s);
+	set_error(&pi->err, "invalid format, CDATA not terminated", pi->str, pi->s);
+	return;
     }
     *end = '\0';
     pi->s = end + 3;
@@ -687,25 +835,50 @@ static char*
 read_quoted_value(PInfo pi) {
     char	*value = 0;
     
-    if ('"' == *pi->s || ('\'' == *pi->s && StrictEffort != pi->options->effort)) {
+    if ('"' == *pi->s || '\'' == *pi->s) {
         char	term = *pi->s;
         
         pi->s++;	/* skip quote character */
         value = pi->s;
         for (; *pi->s != term; pi->s++) {
             if ('\0' == *pi->s) {
-                raise_error("invalid format, document not terminated", pi->str, pi->s);
+                set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+		return 0;
             }
         }
-        *pi->s = '\0'; /* terminate value */
-        pi->s++;	   /* move past quote */
+        *pi->s = '\0';	/* terminate value */
+        pi->s++;	/* move past quote */
     } else if (StrictEffort == pi->options->effort) {
-	raise_error("invalid format, expected a quote character", pi->str, pi->s);
+	set_error(&pi->err, "invalid format, expected a quote character", pi->str, pi->s);
+	return 0;
+    } else if (TolerantEffort == pi->options->effort) {
+        value = pi->s;
+        for (; 1; pi->s++) {
+	    switch (*pi->s) {
+	    case '\0':
+                set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+		return 0;
+	    case ' ':
+	    case '/':
+	    case '>':
+	    case '?': // for instructions
+	    case '\t':
+	    case '\n':
+	    case '\r':
+		pi->last = *pi->s;
+		*pi->s = '\0';	/* terminate value */
+		pi->s++;
+		return value;
+	    default:
+		break;
+            }
+        }
     } else {
         value = pi->s;
         next_white(pi);
 	if ('\0' == *pi->s) {
-	    raise_error("invalid format, document not terminated", pi->str, pi->s);
+	    set_error(&pi->err, "invalid format, document not terminated", pi->str, pi->s);
+	    return 0;
         }
         *pi->s++ = '\0'; /* terminate value */
     }
@@ -752,51 +925,6 @@ read_10_uint64(char *b, uint64_t *up) {
     return b;
 }
 
-/*
-u0000..u007F                00000000000000xxxxxxx  0xxxxxxx
-u0080..u07FF                0000000000yyyyyxxxxxx  110yyyyy 10xxxxxx
-u0800..uD7FF, uE000..uFFFF  00000zzzzyyyyyyxxxxxx  1110zzzz 10yyyyyy 10xxxxxx
-u10000..u10FFFF             uuuzzzzzzyyyyyyxxxxxx  11110uuu 10zzzzzz 10yyyyyy 10xxxxxx
-*/
-static char*
-ucs_to_utf8_chars(char *text, uint64_t u) {
-    int			reading = 0;
-    int			i;
-    unsigned char	c;
-
-    if (u <= 0x000000000000007FULL) {
-	/* 0xxxxxxx */
-	*text++ = (char)u;
-    } else if (u <= 0x00000000000007FFULL) {
-	/* 110yyyyy 10xxxxxx */
-	*text++ = (char)(0x00000000000000C0ULL | (0x000000000000001FULL & (u >> 6)));
-	*text++ = (char)(0x0000000000000080ULL | (0x000000000000003FULL & u));
-    } else if (u <= 0x000000000000D7FFULL || (0x000000000000E000ULL <= u && u <= 0x000000000000FFFFULL)) {
-	/* 1110zzzz 10yyyyyy 10xxxxxx */
-	*text++ = (char)(0x00000000000000E0ULL | (0x000000000000000FULL & (u >> 12)));
-	*text++ = (char)(0x0000000000000080ULL | (0x000000000000003FULL & (u >> 6)));
-	*text++ = (char)(0x0000000000000080ULL | (0x000000000000003FULL & u));
-    } else if (0x0000000000010000ULL <= u && u <= 0x000000000010FFFFULL) {
-	/* 11110uuu 10zzzzzz 10yyyyyy 10xxxxxx */
-	*text++ = (char)(0x00000000000000F0ULL | (0x0000000000000007ULL & (u >> 18)));
-	*text++ = (char)(0x0000000000000080ULL | (0x000000000000003FULL & (u >> 12)));
-	*text++ = (char)(0x0000000000000080ULL | (0x000000000000003FULL & (u >> 6)));
-	*text++ = (char)(0x0000000000000080ULL | (0x000000000000003FULL & u));
-    } else {
-	/* assume it is UTF-8 encoded directly and not UCS */
-	for (i = 56; 0 <= i; i -= 8) {
-	    c = (unsigned char)((u >> i) & 0x00000000000000FFULL);
-	    if (reading) {
-		*text++ = (char)c;
-	    } else if ('\0' != c) {
-		*text++ = (char)c;
-		reading = 1;
-	    }
-	}
-    }
-    return text;
-}
-
 static char*
 read_coded_chars(PInfo pi, char *text) {
     char	*b, buf[32];
@@ -812,7 +940,7 @@ read_coded_chars(PInfo pi, char *text) {
 	}
     }
     if (b > end) {
-	*text++ = *pi->s;
+	*text++ = '&';
     } else if ('#' == *buf) {
 	uint64_t	u = 0;
 	
@@ -823,9 +951,8 @@ read_coded_chars(PInfo pi, char *text) {
 	    b = read_10_uint64(b, &u);
 	}
 	if (0 == b) {
-	    *text++ = *pi->s;
+	    *text++ = '&';
 	} else {
-	    pi->s = s;
 	    if (u <= 0x000000000000007FULL) {
 		*text++ = (char)u;
 #if HAS_PRIVATE_ENCODING
@@ -834,18 +961,25 @@ read_coded_chars(PInfo pi, char *text) {
 #else
 	    } else if (ox_utf8_encoding == pi->options->rb_enc) {
 #endif
-		text = ucs_to_utf8_chars(text, u);
+		text = ox_ucs_to_utf8_chars(text, u);
 #if HAS_PRIVATE_ENCODING
 	    } else if (Qnil == pi->options->rb_enc) {
 #else
 	    } else if (0 == pi->options->rb_enc) {
 #endif
 		pi->options->rb_enc = ox_utf8_encoding;
-		text = ucs_to_utf8_chars(text, u);
+		text = ox_ucs_to_utf8_chars(text, u);
+	    } else if (TolerantEffort == pi->options->effort) {
+		*text++ = '&';
+		return text;
+	    } else if (u <= 0x00000000000000FFULL) {
+		*text++ = (char)u;
 	    } else {
-		/*raise_error("Invalid encoding, need UTF-8 or UTF-16 encoding to parse &#nnnn; character sequences.", pi->str, pi->s); */
-		raise_error("Invalid encoding, need UTF-8 encoding to parse &#nnnn; character sequences.", pi->str, pi->s);
+		/*set_error(&pi->err, "Invalid encoding, need UTF-8 or UTF-16 encoding to parse &#nnnn; character sequences.", pi->str, pi->s); */
+		set_error(&pi->err, "Invalid encoding, need UTF-8 encoding to parse &#nnnn; character sequences.", pi->str, pi->s);
+		return 0;
 	    }
+	    pi->s = s;
 	}
     } else if (0 == strcasecmp(buf, "nbsp;")) {
 	pi->s = s;
@@ -866,7 +1000,7 @@ read_coded_chars(PInfo pi, char *text) {
 	pi->s = s;
 	*text++ = '\'';
     } else {
-	*text++ = *pi->s;
+	*text++ = '&';
     }
     return text;
 }
@@ -884,15 +1018,26 @@ collapse_special(PInfo pi, char *str) {
 	    s++;
 	    if ('#' == *s) {
 		uint64_t	u = 0;
+		char		x;
 
 		s++;
 		if ('x' == *s || 'X' == *s) {
+		    x = *s;
 		    s++;
 		    end = read_hex_uint64(s, &u);
 		} else {
+		    x = '\0';
 		    end = read_10_uint64(s, &u);
 		}
 		if (0 == end) {
+		    if (TolerantEffort == pi->options->effort) {
+			*b++ = '&';
+			*b++ = '#';
+			if ('\0' != x) {
+			    *b++ = x;
+			}
+			continue;
+		    }
 		    return EDOM;
 		}
 		if (u <= 0x000000000000007FULL) {
@@ -903,7 +1048,7 @@ collapse_special(PInfo pi, char *str) {
 #else
 		} else if (ox_utf8_encoding == pi->options->rb_enc) {
 #endif
-		    b = ucs_to_utf8_chars(b, u);
+		    b = ox_ucs_to_utf8_chars(b, u);
 		    /* TBD support UTF-16 */
 #if HAS_PRIVATE_ENCODING
 		} else if (Qnil == pi->options->rb_enc) {
@@ -911,10 +1056,11 @@ collapse_special(PInfo pi, char *str) {
 		} else if (0 == pi->options->rb_enc) {
 #endif
 		    pi->options->rb_enc = ox_utf8_encoding;
-		    b = ucs_to_utf8_chars(b, u);
+		    b = ox_ucs_to_utf8_chars(b, u);
 		} else {
-		    /* raise_error("Invalid encoding, need UTF-8 or UTF-16 encoding to parse &#nnnn; character sequences.", pi->str, pi->s);*/
-		    raise_error("Invalid encoding, need UTF-8 encoding to parse &#nnnn; character sequences.", pi->str, pi->s);
+		    /* set_error(&pi->err, "Invalid encoding, need UTF-8 or UTF-16 encoding to parse &#nnnn; character sequences.", pi->str, pi->s);*/
+		    set_error(&pi->err, "Invalid encoding, need UTF-8 encoding to parse &#nnnn; character sequences.", pi->str, pi->s);
+		    return 0;
 		}
 		s = end + 1;
 	    } else {
@@ -933,6 +1079,9 @@ collapse_special(PInfo pi, char *str) {
 		} else if (0 == strncasecmp(s, "apos;", 5)) {
 		    c = '\'';
 		    s += 5;
+		} else if (TolerantEffort == pi->options->effort) {
+		    *b++ = '&';
+		    continue;
 		} else {
 		    c = '?';
 		    while (';' != *s++) {
